/*************************************************************************/
/* Securetar - a filter to make tar secure                               */
/* Copyright (c) 1998 J. Quade, quade@aachalon.de                        */
/*                                                                       */
/* Securetar is free software; you can redistribute it and/or modify     */
/* it under the terms of the GNU General Public License as published by  */
/* the Free Software Foundation; either version 2, or (at your option)   */
/* any later version.                                                    */
/*                                                                       */
/* Securetar is distributed in the hope that it will be useful,          */
/* but WITHOUT ANY WARRANTY; without even the implied warranty of        */
/* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the          */
/* GNU General Public License for more details.                          */
/*************************************************************************/
#include <stdio.h>
#include <unistd.h>
#include <tar.h>

#define D_BLOCK_SIZE     512
#define D_NAME_LENGTH    100
#define D_UNAME_LENGTH    32
#define D_GNAME_LENGTH    32
#define D_SPARSE_EXT_HDR  21
#define D_SPARSE_IN_HDR    4

#define D_END_OF_ARCHIVE   2

static char *Version = "1.4";

/******************************************************************/
/* The datastructure definition has been directly copied from the */
/* tar sources (version 1.11.8).                                  */
/******************************************************************/
struct sparse {
	char offset[12];
	char numbytes[12];
};

union record {
    unsigned char charptr[D_BLOCK_SIZE];
	struct header {
		char arch_name[D_NAME_LENGTH];
		char mode[8];
		char uid[8];
		char gid[8];
		char size[12];
		char mtime[12];
		char chksum[8];
		char linkflag;
		char arch_linkname[D_NAME_LENGTH];
		char magic[8];
		char uname[D_UNAME_LENGTH];
		char gname[D_GNAME_LENGTH];
		char devmajor[8];
		char devminor[8];

		/* The following fields were added for GNU and are not standard.  */
		char atime[12];
		char ctime[12];
		char offset[12];
		char longnames[4];
		/* Some compilers would insert the pad themselves, so pad was
		   once autoconfigured.  It is simpler to always insert it!  */
		char pad;
		struct sparse sp[D_SPARSE_IN_HDR];
		char isextended;
		char realsize[12];	/* true size of the sparse file */
#if 0
		char ending_blanks[12];	/* number of nulls at the end of the file,
					   if any */
#endif
	} header;

    struct extended_header {
		struct sparse sp[21];
		char isextended;
	}
    ext_hdr;
};

static char *FileTypes[] = {
	"regular file",
	"hard link",
	"symbolic link",
	"character device",
	"block device",
	"directory",
	"named pipe",
	"contignous file"
};

/* Returns the CheckSum difference */
static int ChangeName( unsigned char *Name, int SizeOfName, int Count )
{
	int COld, CNew, i;
	unsigned char Unique[7], *Extension;

	strcpy(Unique,"XXXXXX");
	for( COld=i=0;i<SizeOfName; i++ ) {
		COld += Name[i];
	}
	if( SizeOfName > D_NAME_LENGTH ) {
		Name[D_NAME_LENGTH-1] = '\0';
	}
	if( mktemp(Unique)==NULL )
		perror( Unique );
	/***************************/
	/* Is there any extension? */
	/***************************/
	for( Extension=Name+strlen(Name)-1; Extension>Name && *Extension!='.'
		&& *Extension!='/'; Extension-- )
		;
	if( *Extension=='.' ) {
		sprintf(Name, "t%d.%s%s", Count, Unique, Extension );
	} else {
		sprintf(Name, "t%d.%s", Count, Unique );
	}
	for( CNew=i=0;i<SizeOfName; i++ ) {
		CNew += Name[i];
	}
	return( CNew - COld );
}

static void usage( char *Name )
{
	fprintf(stdout,"%s Version: %s\n", Name, Version );
	fprintf(stdout,"%s removes the pathes of files in a tar ", Name );
	fprintf(stdout,"archive and makes each\nfile name a unique name. ");
	fprintf(stdout,"Links, character devices, block devices and named\n");
	fprintf(stdout,"pipes will be removed from the archive. ");
	fprintf(stdout,"For further information see\n");
	fprintf(stdout,"\thttp://www.aachalon.de/AMaViS/\n");
	fprintf(stdout,"\nusage: %s [options]\n", Name );
	fprintf(stdout,"\tvalid options are:\n");
	fprintf(stdout,"\t-q\tquiet (no output)\n");
	fprintf(stdout,"\nexample:\ncat amavis.tar | %s -q | tar xf -\n\n",Name);
}
static void version( char *Name )
{
	fprintf(stdout,"\n%s %s\n", Name, Version );
}

int main( int argc, char **argv )
{
	int QuietFlag=0;
	union record block;
	int FileNr=0, i, fdin, fdout, ZeroBlockCount=0;
	int StartOfCheckSum, EndOfCheckSum;
	size_t count;
	long FileSize, CheckSum;

	fdin=STDIN_FILENO;
	fdout=STDOUT_FILENO;

	if( argc > 1 ) {
		if( strcmp(argv[1],"-q")==0 ) {
			QuietFlag = 1;
		} else {
			if( strcmp(argv[1],"-v")==0 ) {
				version( argv[0] );
				return( 0 );
			} else {
				usage( argv[0] );
				return( 0 );
				}
		}
	}
	while( (count=read(fdin, &block, sizeof(block))) ) {
		FileNr++;
		/********************************************************/
		/* check for end of archive (2 consecutive zero blocks) */
		/********************************************************/
		if( block.header.arch_name[0] == '\0' ) {
			write( fdout, &block, count );
			if( ++ZeroBlockCount == D_END_OF_ARCHIVE ) {
				while( (count=read(fdin, &block, sizeof(block))) ) {
					write( fdout, &block, count );
				}
				break;
			}
			continue;
		}
		ZeroBlockCount=0;
		/*******************/
		/* Check File Type */
		/*******************/
		if( block.header.linkflag == LNKTYPE
			|| block.header.linkflag == SYMTYPE
			|| block.header.linkflag == CHRTYPE
			|| block.header.linkflag == BLKTYPE
			|| block.header.linkflag == DIRTYPE
			|| block.header.linkflag == FIFOTYPE ) {
			/* Throw it away */
			/* Two "regular" file type exist --> match it to one of them */
			if( block.header.linkflag == 0 )
				block.header.linkflag = '0';
			if( !QuietFlag ) {
				fprintf(stderr,"Removing %s named \"%s\".\n",
					FileTypes[block.header.linkflag-'0'],
					block.header.arch_name);
			}
			continue;
		}
		sscanf( (void *)&block.header.size, "%o", &FileSize );
		/**********************/
		/* Computing Checksum */
		/**********************/
		StartOfCheckSum = (int)&block.header.chksum - (int)&block.header;
		EndOfCheckSum = StartOfCheckSum + sizeof(block.header.chksum);
		for( CheckSum=i=0; i<sizeof(block); i++ ) {
			if( i>=StartOfCheckSum && i<EndOfCheckSum ) {
				CheckSum += (unsigned char)' ';
			} else {
				CheckSum += block.charptr[i];
			}
		}
		if( !QuietFlag ) {
			fprintf(stderr,"changed name from \"%s\"",
				block.header.arch_name );
		}
		CheckSum += ChangeName(block.header.arch_name,
			sizeof(block.header.arch_name), FileNr );
		if( !QuietFlag ) {
			fprintf(stderr," to \"%s\"\n", block.header.arch_name );
		}
		sprintf(block.header.chksum,"%6o",CheckSum );
		/***********************************/
		/* Write out the (new) Headerblock */
		/***********************************/
		write( fdout, &block, count );
		/************************************/
		/* Write out the corresponding data */
		/************************************/
		while( FileSize>0 ) {
			FileSize-=(long)count=read(fdin,&block,sizeof(block));
			write( fdout, &block, count );
		}
	}
	return( 0 );
}
